#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Implementation for evolutionary multitasking (EM)"""
from math import tanh
from operator import itemgetter
from collections import defaultdict
from ase.ga.population import Population
from ase.ga.convergence import Convergence
from ase.ga import get_raw_score
import numpy as np


class MultitaskPopulation(Population):                                      
    """Different tasks are assigned to different niches. The candidates
    are ranked according to the effective fitness, given by the shortest 
    distance between the raw score of the marked niche and the upper 
    envelope after adding the individual. The raw score is given by the 
    fitness gain in the maximum-gained niche. **The raw scores of each 
    configuration for all tasks must be provided as a Numpy array in
    atoms.info['data']['raw_scores']**. After providing the raw scores, 
    **the effective score of each configuration is automatically 
    calculated and stored in atoms.info['key_value_pairs']['raw_score']**.
    The niches that benefit the most from each configuration are stored in
    atoms.info['data']['niches'], and the dominating niche is stored in
    atoms.info['key_value_pairs']['dominating_niche'].

    Parameters
    ----------
    num_tasks: int
        The number of tasks.

    exp_function: bool, default True
        If True use an exponential function for ranking the fitness.
        If False use the same as in Population.

    exp_prefactor: float, default 0.5
        The prefactor used in the exponential fitness scaling function.
    """

    def __init__(self, data_connection, population_size, num_tasks,
                 comparator=None, logfile=None, use_extinct=False,
                 exp_function=True, exp_prefactor=0.5, rng=np.random):
        self.exp_function = exp_function
        self.exp_prefactor = exp_prefactor
        self.vf = lambda x: x.info['key_value_pairs']['dominating_niche']
        # The current fitness is set at each update of the population
        self.current_fitness = None

        Population.__init__(self, data_connection, population_size,
                            comparator, logfile, use_extinct, rng=rng)
        self.max_scores = np.full(num_tasks, np.NINF, dtype=float)
        self.dominating_niches = np.full(num_tasks, -1, dtype=int)
        self.rep_no_gain = 0

    def get_rank(self, candidates):
        rank = np.array([-1] * len(candidates))
        # Remember the order when decreasing rank later
        order = dict((candidates[i].info['key_value_pairs']['gaid'], i)
                     for i in range(len(candidates)))

        # Group candidates in niches according to the variable
        # function vf and also sort them according to raw score
        self.set_vf_dict(candidates, key=get_raw_score, reverse=True)
        # Decrease the rank of the not best candidates in each niche
        for vf, li in self.vf_dict.items():
            for i, c in enumerate(li):
                rank[order[c.info['key_value_pairs']['gaid']]] -= i

        return rank

    def set_vf_dict(self, candidates, **sort_arguments):
        d = defaultdict(list)
        for c in candidates:
            d[self.vf(c)].append(c)
        if sort_arguments:
            for cl in d.values():
                cl.sort(**sort_arguments)
        self.vf_dict = d

    def __get_fitness__(self, candidates):
        expf = self.exp_function
        rfit = self.get_rank(candidates)

        if not expf:
            rmax = max(rfit)
            rmin = min(rfit)
            T = rmin - rmax
            # If using obj_rank probability, must have non-zero T val.
            # pop_size must be greater than number of permutations.
            # We test for this here
            msg = "Equal fitness for best and worst candidate in the "
            msg += "population! Fitness scaling is impossible! "
            msg += "Try with a larger population."
            assert T != 0., msg
            return 0.5 * (1. - np.tanh(2. * (rfit - rmax) / T - 1.))
        else:
            return self.exp_prefactor ** (-rfit - 1)

    def update(self, new_cand):
        """The update method in Population will add to the end of
        the population, that can't be used here since the fitness
        will potentially change for all candidates when new are added,
        therefore just recalc the population every time. New candidates
        are required (must not be added before calling this method).
        The maximum gain dynamic niching (MGDN) algorithm is executed.
        """

        # Update the upper envelope
        prev_max_scores = self.max_scores.copy()
        gained_ids = []
        for i, a in enumerate(new_cand):
            scores = a.info['data']['raw_scores']
            gained_niches = np.argwhere(scores > self.max_scores)
            if gained_niches.size != 0:
                self.max_scores[gained_niches] = scores[gained_niches]
                gained_ids.append(i)

        # Update the array that records the niche dominating other gained niches
        # with the requirements of: 1. contributes to the updated upper envelope;
        # 2. maximum in gain compared to the previous upper envelope
        first_generation = np.any(prev_max_scores == np.NINF)
        for i in gained_ids:
            scores = new_cand[i].info['data']['raw_scores']
            maxed_niches = np.argwhere(scores == self.max_scores)
            if maxed_niches.size != 0:
                if first_generation:
                    dominating_niche = int(max(maxed_niches, key=lambda x: scores[x]))
                else:
                    dominating_niche = int(max(maxed_niches, key=lambda x:
                                               scores[x] - prev_max_scores[x]))
                self.dominating_niches[maxed_niches] = dominating_niche

        # Caculate the effective fitness and assign a niche for each new candidate
        for i in range(len(new_cand)):
            scores = new_cand[i].info['data']['raw_scores']
            min_loss_niche = np.argmax(scores - self.max_scores)
            dominating_niche = self.dominating_niches[min_loss_niche]
            f_eff = float(np.around(scores[min_loss_niche] - 
                          self.max_scores[min_loss_niche], 8))
            new_cand[i].info['key_value_pairs']['raw_score'] = f_eff
            new_cand[i].info['key_value_pairs']['dominating_niche'] = dominating_niche
            new_cand[i].info['data']['niches'] = np.argwhere(
                self.dominating_niches==dominating_niche).flatten() 

        # Update the fitness of all previously-relaxed candidates if fitness     
        # is gained at any niche from the new generation (niche migration)
        updated_cand = []
        if gained_ids and len(self.pop) > 0:
            # Update the database        
            prev_cand = self.dc.get_all_relaxed_candidates()
            del_ids = [] 
            for a in prev_cand:
                scores = a.info['data']['raw_scores']
                min_loss_niche = np.argmax(scores - self.max_scores)
                dominating_niche = self.dominating_niches[min_loss_niche]
                f_eff = float(np.around(scores[min_loss_niche] - 
                              self.max_scores[min_loss_niche], 8))
                a.info['key_value_pairs']['raw_score'] = f_eff
                a.info['key_value_pairs']['dominating_niche'] = dominating_niche
                a.info['data']['niches'] = np.argwhere(
                    self.dominating_niches==dominating_niche).flatten() 
                updated_cand.append(a)
                gaid = a.info['confid']
                del_ids.append(gaid)
            self.dc.c.delete(del_ids)
            self.rep_no_gain = 0
        else:
            self.rep_no_gain += 1

        self.dc.add_more_relaxed_candidates(updated_cand + new_cand)
        self.pop = []            
        self.__initialize_pop__()
        
        self._write_log()

    def __initialize_pop__(self):
        # Get all relaxed candidates from the database
        ue = self.use_extinct
        all_cand = self.dc.get_all_relaxed_candidates(use_extinct=ue)
        all_cand.sort(key=get_raw_score, reverse=True)

        if len(all_cand) > 0:
            fitf = self.__get_fitness__(all_cand)
            all_sorted = list(zip(fitf, all_cand))
            all_sorted.sort(key=itemgetter(0), reverse=True)
            sort_cand = []
            for _, t2 in all_sorted:
                sort_cand.append(t2)
            all_sorted = sort_cand

            # Fill up the population with the self.pop_size most stable
            # unique candidates.
            i = 0
            while i < len(all_sorted) and len(self.pop) < self.pop_size:
                c = all_sorted[i]
                c_vf = self.vf(c)
                i += 1
                eq = False
                for a in self.pop:
                    a_vf = self.vf(a)
                    # Only run comparator if the variable_function (self.vf)
                    # returns the same. If it returns something different the
                    # candidates are inherently different.
                    # This is done to speed up.
                    if a_vf == c_vf:
                        if self.comparator.looks_like(a, c):
                            eq = True
                            break
                if not eq:
                    self.pop.append(c)
        self.current_fitness = self.__get_fitness__(self.pop)
        self.all_cand = all_cand

    def get_two_candidates(self):
        """Returns two candidates for pairing employing the
        roulete wheel selection scheme described in
        R.L. Johnston Dalton Transactions,
        Vol. 22, No. 22. (2003), pp. 4193-4207
        """

        if len(self.pop) < 2:
            self.update()

        if len(self.pop) < 2:
            return None

        # Use saved fitness
        fit = self.current_fitness
        fmax = max(fit)
        c1 = self.pop[0]
        c2 = self.pop[0]
        while c1.info['confid'] == c2.info['confid']:
            nnf = True
            while nnf:
                t = self.rng.randint(len(self.pop))
                if fit[t] > self.rng.rand() * fmax:
                    c1 = self.pop[t]
                    nnf = False
            nnf = True
            while nnf:
                t = self.rng.randint(len(self.pop))
                if fit[t] > self.rng.rand() * fmax:
                    c2 = self.pop[t]
                    nnf = False

        return (c1.copy(), c2.copy())


class MultitaskGenerationRepetitionConvergence(Convergence):                      
    """Returns True if the latest finished population is stagnated 
    (no fitness gain in any task) for number_of_generations.

    Parameters
    ----------
    number_of_generations: int
        How many generations need to be equal before convergence.

    max_generations: int, default indefinte
        The maximum number of generations the GA is allowed to run.
    """

    def __init__(self, population_instance, 
                 number_of_generations,
                 max_generations=100000000):
        Convergence.__init__(self, population_instance)
        self.numgens = number_of_generations
        self.maxgen = max_generations

    def converged(self):
        size = self.pop.pop_size
        cur_gen_num = self.pop.dc.get_generation_number(size)

        if cur_gen_num >= self.maxgen:
            return True

        if cur_gen_num <= 1:
            return False

        if self.pop.rep_no_gain >= self.numgens:
            return True

        return False




#!/usr/bin/env python
# -*- coding: utf-8 -*-
from .settings import (adsorbate_elements,
                       site_heights,
                       adsorbate_formulas)
from .adsorption_sites import (ClusterAdsorptionSites, 
                               SlabAdsorptionSites)
from .utilities import (neighbor_shell_list, 
                        get_adj_matrix,
                        get_mic)
from ase.data import atomic_numbers
from ase.geometry import find_mic
from ase.formula import Formula
from collections import defaultdict, Counter
from operator import attrgetter
from copy import deepcopy
import networkx as nx
import numpy as np
import random


class ClusterAdsorbateCoverage(object):
    """Child class of `ClusterAdsorptionSites` for identifying adsorbate 
    coverage on a nanoparticle. Support common nanoparticle shapes 
    including: Mackay icosahedron, (truncated) octahedron and (Marks) 
    decahedron.

    The information of each occupied site stored in the dictionary is 
    updated with the following new keys:

    **'occupied'**: 1 if the site is occupied, otherwise 0.
 
    **'adsorbate'**: the name of the adsorbate that occupies this site.
 
    **'adsorbate_indices'**: the indices of the adosorbate atoms that occupy
    this site. If the adsorbate is multidentate, these atoms might
    occupy multiple sites.
 
    **'bonding_index'**: the index of the atom that directly bonds to the
    site (closest to the site).
 
    **'fragment'**: the name of the fragment that occupies this site. Useful
    for multidentate species.
 
    **'fragment_indices'**: the indices of the fragment atoms that occupy
    this site. Useful for multidentate species.
 
    **'bond_length'**: the distance between the bonding atom and the site.
 
    **'dentate'**: dentate number.
 
    **'label'**: the updated label with the name of the occupying adsorbate
    if label_occupied_sites is set to True.

    Parameters
    ----------
    atoms : ase.Atoms object
        The atoms object must be a non-periodic nanoparticle with at 
        least one adsorbate attached to it. Accept any ase.Atoms object. 
        No need to be built-in.

    adsorption_sites : acat.adsorption_sites.ClusterAdsorptionSites object, \
        default None
        `ClusterAdsorptionSites` object of the nanoparticle. Initialize a 
        `ClusterAdsorptionSites` object if not specified.
        If this is not provided, the arguments for identifying adsorption 
        sites can still be passed in by **kwargs.

    subtract_heights : dict, default None
        A dictionary that contains the height to be subtracted from the bond 
        length when allocating a type of site to an adsorbate. Default is to 
        allocate the site that is closest to the adsorbate's binding atom 
        without subtracting height. Useful for ensuring the allocated site 
        for each adsorbate is consistent with the site to which the adsorbate
        was added.                                                           

    label_occupied_sites : bool, default False
        Whether to assign a label to the occupied each site. The string 
        of the occupying adsorbate is concatentated to the numerical 
        label that represents the occupied site.

    dmax : float, default 2.5
        The maximum bond length (in Angstrom) between an atom and its
        nearest site to be considered as the atom being bound to the site.    

    Example
    -------
    The following example illustrates the most important use of a
    `ClusterAdsorbateCoverage` object - getting occupied adsorption sites:

        >>> from acat.adsorption_sites import ClusterAdsorptionSites
        >>> from acat.adsorbate_coverage import ClusterAdsorbateCoverage
        >>> from acat.build import add_adsorbate_to_site
        >>> from ase.cluster import Octahedron
        >>> atoms = Octahedron('Ni', length=7, cutoff=2)
        >>> for atom in atoms:
        ...     if atom.index % 2 == 0:
        ...         atom.symbol = 'Pt'
        >>> atoms.center(vacuum=5.)
        >>> cas = ClusterAdsorptionSites(atoms, composition_effect=True,
        ...                              surrogate_metal='Ni') 
        >>> sites = cas.get_sites()
        >>> for s in sites:
        ...     if s['site'] == 'fcc':
        ...         add_adsorbate_to_site(atoms, adsorbate='CO', site=s)
        >>> cac = ClusterAdsorbateCoverage(atoms, adsorption_sites=cas,
        ...                                label_occupied_sites=True)
        >>> occupied_sites = cac.get_sites(occupied_only=True)
        >>> print(occupied_sites[0])

    Output:

    .. code-block:: python

        {'site': 'fcc', 'surface': 'fcc111', 
         'position': array([ 6.17333333,  7.93333333, 11.45333333]), 
         'normal': array([-0.57735027, -0.57735027, -0.57735027]), 
         'indices': (0, 2, 4), 'composition': 'PtPtPt', 
         'subsurf_index': None, 'subsurf_element': None, 'label': '21CO', 
         'bonding_index': 201, 'bond_length': 1.3, 'adsorbate': 'CO', 
         'fragment': 'CO', 'adsorbate_indices': (201, 202), 'occupied': 1, 
         'dentate': 1, 'fragment_indices': (201, 202)}

    """

    def __init__(self, atoms, 
                 adsorption_sites=None, 
                 subtract_heights=None,
                 label_occupied_sites=False,
                 dmax=2.5, **kwargs):

        atoms = atoms.copy()
        for dim in range(3):
            if np.linalg.norm(atoms.cell[dim]) == 0:
                atoms.cell[dim][dim] = np.ptp(atoms.positions[:, dim]) + 10.

        self.atoms = atoms 
        self.positions = atoms.positions
        self.symbols = atoms.symbols
        self.ads_ids = [a.index for a in atoms if 
                        a.symbol in adsorbate_elements]
        if self.ads_ids:
            self.ads_atoms = atoms[self.ads_ids]
        else:
            self.ads_atoms = None
        self.cell = atoms.cell
        self.pbc = atoms.pbc

        self.subtract_heights = subtract_heights
        if subtract_heights is not None:
            self.subtract_heights = site_heights
            for k, v in subtract_heights.items():
                self.subtract_heights[k] = v      
        self.label_occupied_sites = label_occupied_sites
        self.dmax = dmax
        self.kwargs = {'allow_6fold': False, 'composition_effect': False,
                       'ignore_bridge_sites': False, 'label_sites': False} 
        self.kwargs.update(kwargs)
        if self.ads_ids:
            self.make_ads_neighbor_list()
            self.ads_adj_matrix = self.get_ads_connectivity() 
            self.identify_adsorbates()
        else:
            self.ads_list = []

        if adsorption_sites:
            cas = adsorption_sites
            for k in self.kwargs.keys():
                self.kwargs[k] = attrgetter(k)(cas)
        else:
            cas = ClusterAdsorptionSites(atoms, **self.kwargs)    
        self.__dict__.update(self.kwargs)
        self.cas = cas
        self.slab_ids = cas.indices
        self.metals = cas.metals
        self.surf_ids = cas.surf_ids
        self.label_dict = cas.label_dict 
        self.hetero_site_list = deepcopy(cas.site_list)

        self.label_list = ['0'] * len(self.hetero_site_list)
        self.populate_hetero_site_list()

        if self.ads_ids:
            self.labels = self.get_occupied_labels()
        else:
            self.labels = []

    def identify_adsorbates(self):

        G = nx.Graph()
        adscm = self.ads_adj_matrix

        # Cut all intermolecular H-H bonds except intramolecular               
        # H-H bonds in e.g. H2
        hids = [a.index for a in self.ads_atoms if a.symbol == 'H']
        for hi in hids:
            conns = np.where(adscm[hi] == 1)[0]
            hconns = [i for i in conns if self.ads_atoms.symbols[i] == 'H']
            if hconns and len(conns) > 1:
                adscm[hi,hconns] = 0
      
        if adscm.size != 0:
            np.fill_diagonal(adscm, 1)
            rows, cols = np.where(adscm == 1)

            edges = zip([self.ads_ids[row] for row in rows.tolist()], 
                        [self.ads_ids[col] for col in cols.tolist()])
            G.add_edges_from(edges)                        
            SG = (G.subgraph(c) for c in nx.connected_components(G))
 
            adsorbates = []
            for sg in SG:
                nodes = sorted(list(sg.nodes))
                adsorbates += [nodes]
        else:
            adsorbates = [self.ads_ids]

        self.ads_list = adsorbates

    def get_hetero_connectivity(self):
        """Get the adjacency matrix of slab + adsorbates."""

        nbslist = neighbor_shell_list(self.atoms, 0.3, neighbor_number=1)
        return get_adj_matrix(nbslist)                          

    def get_ads_connectivity(self):
        """Get the adjacency matrix for adsorbate atoms."""

        return get_adj_matrix(self.ads_nblist) 

    def get_site_connectivity(self):
        """Get the adjacency matrix for adsorption sites."""

        sl = self.hetero_site_list
        conn_mat = []
        for i, sti in enumerate(sl):
            conn_x = []
            for j, stj in enumerate(sl): 
                overlap = len(set(sti['indices']).intersection(stj['indices']))
                if i == j:
                    conn_x.append(0.)
                elif overlap > 0:
                    if self.allow_6fold: 
                        if '6fold' in [sti['site'], stj['site']]: 
                            if overlap == 3:                            
                                conn_x.append(1.)
                            else:
                                conn_x.append(0.)
                        else:
                            conn_x.append(1.)
                    else:
                        conn_x.append(1.)                     
                else:
                    conn_x.append(0.)
            conn_mat.append(conn_x)   

        return np.asarray(conn_mat) 

    def populate_hetero_site_list(self):
        """Find all the occupied sites, identify the adsorbate coverage
        of those sites and collect in a heterogeneous site list."""

        hsl = self.hetero_site_list
        ll = self.label_list
        ads_list = self.ads_list
        ndentate_dict = {} 

        for adsid in self.ads_ids:            
            if self.symbols[adsid] == 'H':
                if [adsid] not in ads_list:
                    rest = [s for x in ads_list for s in x 
                            if (adsid in x and s != adsid)]
                    if not (self.symbols[rest[0]] == 'H' and len(rest) == 1):
                        continue

            adspos = self.positions[adsid]
            if self.subtract_heights is not None:
                dls = np.linalg.norm(np.asarray([s['position'] + s['normal'] * 
                                     self.subtract_heights[s['site']] - adspos 
                                     for s in hsl]), axis=1)
                stid = np.argmin(dls)
                st = hsl[stid]
                bl = np.linalg.norm(st['position'] - adspos)
            else:
                bls = np.linalg.norm(np.asarray([s['position'] - 
                                     adspos for s in hsl]), axis=1)
                stid = np.argmin(bls)
                st, bl = hsl[stid], bls[stid]
            bl = round(bl, 8)
            if bl > self.dmax:
                continue

            adsids = next((l for l in ads_list if adsid in l), None)
            adsi = tuple(sorted(adsids))

            if 'occupied' in st:
                if bl >= st['bond_length']:
                    continue
                elif self.symbols[adsid] != 'H':
                    if adsi in ndentate_dict:
                        ndentate_dict[adsi] -= 1 
            st['bonding_index'] = adsid
            st['bond_length'] = bl

            symbols = ''.join(list(Formula(str(self.symbols[adsids]))))
            adssym = next((k for k, v in adsorbate_formulas.items() 
                           if v == symbols), None)
            if adssym is None:
                adssym = next((k for k, v in adsorbate_formulas.items()
                               if sorted(v) == sorted(symbols)), symbols)

            st['adsorbate'] = adssym
            st['fragment'] = adssym
            st['adsorbate_indices'] = adsi 
            if adsi in ndentate_dict:
                ndentate_dict[adsi] += 1
            else:
                ndentate_dict[adsi] = 1
            st['occupied'] = 1            

        # Get dentate numbers and coverage  
        self.n_occupied, n_surf_occupied, self.n_subsurf_occupied = 0, 0, 0
        for st in hsl:
            if 'occupied' not in st:
                st['bonding_index'] = st['bond_length'] = None
                st['adsorbate'] = st['fragment'] = None
                st['adsorbate_indices'] = None
                st['occupied'] = st['dentate'] = 0
                st['fragment_indices'] = None
                st['label'] = 0
                continue
            self.n_occupied += 1
            if st['site'] == '6fold':
                self.n_subsurf_occupied += 1
            else:
                n_surf_occupied += 1
            adsi = st['adsorbate_indices']
            if adsi in ndentate_dict:              
                st['dentate'] = ndentate_dict[adsi]
            else:
                st['dentate'] = 0
        self.coverage = n_surf_occupied / len(self.surf_ids)

        # Identify bidentate fragments and assign labels 
        self.multidentate_fragments = []
        self.monodentate_adsorbate_list = []
        self.multidentate_labels = []
        multidentate_adsorbate_dict = {}
        for j, st in enumerate(hsl):
            if st['occupied']:
                adssym = st['adsorbate']
                if st['dentate'] > 1:
                    bondid = st['bonding_index']
                    bondsym = self.symbols[bondid]
                    conns = [self.ads_ids[k] for k in np.where(self.ads_adj_matrix[
                             self.ads_ids.index(bondid)] == 1)[0].tolist()]
                    hnnlen = len([i for i in conns if self.atoms[i].symbol == 'H'])
                    fsym = self.atoms[bondid].symbol
                    if hnnlen == 1:
                        fsym += 'H'
                    elif hnnlen > 1:
                        fsym += 'H{}'.format(hnnlen)
                    st['fragment'] = fsym
                    flen = hnnlen + 1
                    adsids = st['adsorbate_indices']
                    ibond = adsids.index(bondid)
                    fsi = adsids[ibond:ibond+flen]
                    st['fragment_indices'] = fsi
                    if adsids not in multidentate_adsorbate_dict:
                        multidentate_adsorbate_dict[adsids] = adssym
                else:
                    st['fragment_indices'] = st['adsorbate_indices']
                    self.monodentate_adsorbate_list.append(adssym)

                if self.label_occupied_sites:
                    if st['label'] is None:
                        signature = [st['site'], st['surface']] 
                        if self.composition_effect:
                            signature.append(st['composition'])
                        stlab = self.label_dict['|'.join(signature)]
                    else:
                        stlab = st['label']                         
                    label = str(stlab) + st['fragment']
                    st['label'] = label
                    ll[j] = label
                    if st['dentate'] > 1:                    
                        self.multidentate_fragments.append(label)
                        if bondid == adsids[0]:
                            mdlabel = str(stlab) + adssym
                            self.multidentate_labels.append(mdlabel)

        self.multidentate_adsorbate_list = list(multidentate_adsorbate_dict.values())
        self.adsorbate_list = self.monodentate_adsorbate_list + \
                              self.multidentate_adsorbate_list 

    def get_site(self, indices=None):
        """Get information of an adsorption site.
        
        Parameters
        ----------
        indices : list or tuple, default None
            The indices of the atoms that contribute to the site. 
            Return a random site if the indices are not provided.        

        """

        if indices is None:
            st = random.choice(self.hetero_site_list)
        else:
            indices = indices if is_list_or_tuple(indices) else [indices]
            indices = tuple(sorted(indices))
            st = next((s for s in self.hetero_site_list if 
                       s['indices'] == indices), None)
        return st

    def get_sites(self, occupied_only=False):
        """Get information of all sites.
                                                                     
        Parameters                                                   
        ----------                                                   
        occupied_only : bool, default False
            Whether to only return occupied sites.

        """                                                          

        all_sites = self.hetero_site_list
        if occupied_only:
            all_sites = [s for s in all_sites if s['occupied']]
        return all_sites

    def get_adsorbates(self, adsorbate_species=None):
        """Get a list of tuples that contains each adsorbate (string)
        and the corresponding adsorbate indices (tuple) in ascending
        order.

        Parameters
        ----------
        adsorbate_species : list of strs, default None
            The available adsorbate species. If the adsorbate is not
            one of the available species, return the fragment instead.

        """

        adsorbates = []
        adsorbate_ids = set()
        for st in self.hetero_site_list:
            if st['occupied']:
                if adsorbate_species is not None:
                    if st['adsorbate'] not in adsorbate_species:
                        fragi = st['fragment_indices']
                        adsorbates.append((st['fragment'], fragi))
                        adsorbate_ids.update(fragi)
                        continue
                adsi = st['adsorbate_indices']
                if not set(adsi).issubset(adsorbate_ids):
                    adsorbates.append((st['adsorbate'], adsi))
                    adsorbate_ids.update(adsi)

        return sorted(adsorbates, key=lambda x: x[1])

    def get_fragments(self):
        """Get a list of tuples that contains each fragment (string)
        and the corresponding fragment indices (tuple) in ascending
        order."""

        fragments = []
        for st in self.hetero_site_list:
            if st['occupied']:
                fragi = st['fragment_indices']
                fragments.append((st['fragment'], fragi))

        return sorted(fragments, key=lambda x: x[1])        

    def make_ads_neighbor_list(self, dx=.2, neighbor_number=1):
        self.ads_nblist = neighbor_shell_list(self.ads_atoms, dx, 
                                              neighbor_number, mic=False)

    def get_occupied_labels(self, fragmentation=True):
        """Get a list of labels of all occupied sites. The label consists
        of a numerical part that represents site, and a character part
        that represents the occupying adsorbate.

        Parameters
        ----------
        fragmentation : bool, default True
            Whether to cut multidentate species into fragments. This ensures 
            that multidentate species with different orientations have 
            different labels.

        """

        if not self.label_occupied_sites:
            return self.atoms[self.ads_ids].get_chemical_formula(mode='hill')

        ll = self.label_list
        labs = [lab for lab in ll if lab != '0']
        if not fragmentation:
            mf = self.multidentate_fragments
            mdlabs = self.multidentate_labels
            c1, c2 = Counter(labs), Counter(mf)
            diff = list((c1 - c2).elements())
            labs = diff + mdlabs                   

        return sorted(labs)

    def get_graph(self, atom_wise=False,
                  fragmentation=True, 
                  subsurf_effect=False, 
                  full_effect=False,
                  return_adj_matrix=False,
                  connect_dentates=True,
                  dx=0.5):                                         
        """Get the graph representation of the nanoparticle with adsorbates.

        Parameters
        ----------
        atom_wise : bool, default False
            Whether to treat each adsorbate as an atom-wise subgraph. If not, 
            treat each adsorbate fragment as a unity (molecule-wise).

        fragmentation : bool, default True
            Whether to cut multidentate species into fragments. This ensures 
            that multidentate species with different orientations have 
            different graphs.

        subsurf_effect : bool, default False
            Take subsurface atoms into consideration when genearting graph. 

        full_effect : bool, default False
            Take the whole catalyst into consideration when generating graph.

        return_adj_matrix : bool, default False
            Whether to return adjacency matrix instead of the networkx.Graph 
            object.

        connect_dentates : bool, default True
            Whether to add edges between the fragments of multidentate species.

        dx : float, default 0.5
            Buffer to calculate nearest neighbor pairs. Only relevent when
            atom_wise=True.

        """

        # Molecule-wise
        if not atom_wise:
            hsl = self.hetero_site_list
            hcm = self.cas.get_connectivity().copy()
            if full_effect:
                surf_ids = self.slab_ids
            else:
                if subsurf_effect:
                    surf_ids = self.surf_ids + self.cas.get_subsurface()
                else:
                    surf_ids = self.surf_ids
            surfhcm = hcm[surf_ids]
            symbols = self.symbols[surf_ids]
            nrows, ncols = surfhcm.shape       
 
            newrows, frag_list, adsi_list = [], [], []
            multi_conns = defaultdict(list)
            ohsl = [st for st in hsl if st['occupied']]
            for st in sorted(ohsl, key=lambda x: x['fragment_indices']):
                adsi = st['adsorbate_indices']
                if not fragmentation and st['dentate'] > 1: 
                    if st['bonding_index'] != adsi[0]:
                        continue
                si = st['indices']
                newrow = np.zeros(ncols)
                newrow[list(si)] = 1
                newrows.append(newrow)
 
                if fragmentation:                
                    frag_list.append(st['fragment'])
                    if st['dentate'] > 1 and connect_dentates:
                        bondid = st['bonding_index']
                        all_conns = [self.ads_ids[j] for j in np.where(self.ads_adj_matrix[
                                     self.ads_ids.index(bondid)] == 1)[0].tolist()]
                        conns = [c for c in all_conns if c not in st['fragment_indices']]
                        i = len(newrows) - 1
                        if i not in multi_conns:
                            multi_conns[bondid].append([i])
                        for k, v in multi_conns.items():
                            if k in conns:
                                multi_conns[k].append([i, v[0][0]])
                else:
                    frag_list.append(st['adsorbate'])
 
            links = []
            if newrows:
                surfhcm = np.vstack((surfhcm, np.asarray(newrows)))
                if multi_conns:
                    links = [sorted(c) for cs in multi_conns.values() for c in cs if len(c) > 1]
 
            shcm = surfhcm[:,surf_ids]
            if return_adj_matrix:
                if newrows:
                    dd = len(newrows)
                    small_mat = np.zeros((dd, dd))
                    if links:
                        for (i, j) in links:
                            small_mat[i,j] = 1
                            small_mat[j,i] = 1
                    return np.hstack((shcm, np.vstack((shcm[-dd:].T, small_mat))))
                else:
                    return shcm
 
            G = nx.Graph()               
            # Add nodes from fragment list
            G.add_nodes_from([(i, {'symbol': symbols[i]}) 
                               for i in range(nrows)] + 
                             [(j + nrows, {'symbol': frag_list[j]})
                               for j in range(len(frag_list))])
 
            # Add edges from surface adjacency matrix
            shcm = shcm * np.tri(*shcm.shape, k=-1)
            rows, cols = np.where(shcm == 1)
            edges = zip(rows.tolist(), cols.tolist())
            G.add_edges_from(edges)
            if links:
                for (i, j) in links:
                    G.add_edge(i + len(surf_ids), j + len(surf_ids))
        # Atom-wise
        else:
            nblist = neighbor_shell_list(self.atoms, dx=dx, 
                                         neighbor_number=1, mic=False)
            cm = get_adj_matrix(nblist)
            if full_effect:
                surf_ids = self.slab_ids
            else:
                if subsurf_effect:
                    surf_ids = self.surf_ids + self.subsurf_ids
                else:
                    surf_ids = self.surf_ids
            surf_ads_ids = sorted(surf_ids + self.ads_ids)
            shcm = cm[surf_ads_ids]
            symbols = self.symbols[surf_ads_ids]
            if return_adj_matrix:
                return shcm
 
            G = nx.Graph()                                                  
            G.add_nodes_from([(i, {'symbol': symbols[i]}) 
                              for i in range(len(symbols))])
            rows, cols = np.where(shcm == 1)
            edges = zip(rows.tolist(), cols.tolist())
            G.add_edges_from(edges)

        return G

    def get_coverage(self):
        """Get the adsorbate coverage (ML) of the surface."""

        return self.coverage

    def get_subsurf_coverage(self):
        """Get the adsorbate coverage (ML) of the subsurface."""

        nsubsurf = len(self.cas.get_subsurface())
        return self.n_subsurf_occupied / nsubsurf


class SlabAdsorbateCoverage(object):
    """Child class of `SlabAdsorptionSites` for identifying adsorbate 
    coverage on a surface slab. Support 20 common surfaces: fcc100, 
    fcc111, fcc110, fcc211, fcc221, fcc311, fcc322, fcc331, fcc332, 
    bcc100, bcc111, bcc110, bcc210, bcc211, bcc310, hcp0001, 
    hcp10m10t, hcp10m10h, hcp10m11, hcp10m12.

    The information of each occupied site stored in the dictionary is 
    updated with the following new keys:

    **'occupied'**: 1 if the site is occupied, otherwise 0.

    **'adsorbate'**: the name of the adsorbate that occupies this site.

    **'adsorbate_indices'**: the indices of the adosorbate atoms that occupy
    this site. If the adsorbate is multidentate, these atoms might
    occupy multiple sites.

    **'bonding_index'**: the index of the atom that directly bonds to the
    site (closest to the site).

    **'fragment'**: the name of the fragment that occupies this site. Useful
    for multidentate species.

    **'fragment_indices'**: the indices of the fragment atoms that occupy
    this site. Useful for multidentate species.

    **'bond_length'**: the distance between the bonding atom and the site.

    **'dentate'**: dentate number.

    **'label'**: the updated label with the name of the occupying adsorbate
    if label_occupied_sites is set to True.

    Parameters
    ----------
    atoms : ase.Atoms object
        The atoms object must be a non-periodic nanoparticle with at 
        least one adsorbate attached to it. Accept any ase.Atoms object. 
        No need to be built-in.

    adsorption_sites : acat.adsorption_sites.SlabAdsorptionSites object, \
        default None
        `SlabAdsorptionSites` object of the nanoparticle. Initialize a         
        `SlabAdsorptionSites` object if not specified.
        If this is not provided, the arguments for identifying adsorption 
        sites can still be passed in by **kwargs.

    subtract_heights : dict, default None
        A dictionary that contains the height to be subtracted from the bond 
        length when allocating a type of site to an adsorbate. Default is to 
        allocate the site that is closest to the adsorbate's binding atom 
        without subtracting height. Useful for ensuring the allocated site 
        for each adsorbate is consistent with the site to which the adsorbate 
        was added.                                                             

    label_occupied_sites : bool, default False
        Whether to assign a label to the occupied each site. The string 
        of the occupying adsorbate is concatentated to the numerical 
        label that represents the occupied site.

    dmax : float, default 2.5
        The maximum bond length (in Angstrom) between an atom and its
        nearest site to be considered as the atom being bound to the site.

    Example
    -------
    The following example illustrates the most important use of a
    `SlabAdsorbateCoverage` object - getting occupied adsorption sites:

        >>> from acat.adsorption_sites import SlabAdsorptionSites
        >>> from acat.adsorbate_coverage import SlabAdsorbateCoverage
        >>> from acat.build import add_adsorbate
        >>> from ase.build import fcc211
        >>> atoms = fcc211('Cu', (3, 3, 4), vacuum=5.)
        >>> for atom in atoms:
        ...     if atom.index % 2 == 0:
        ...         atom.symbol = 'Au'
        >>> atoms.center()
        >>> add_adsorbate(atoms, adsorbate='CH3OH', surface='fcc211',
        ...               indices=(5, 7, 8), surrogate_metal='Cu')
        >>> sac = SlabAdsorbateCoverage(atoms, surface='fcc211', 
        ...                             label_occupied_sites=True)
        >>> occupied_sites = sac.get_sites(occupied_only=True)
        >>> print(occupied_sites)

    Output:

    .. code-block:: python

        [{'site': 'bridge', 'surface': 'fcc211', 'morphology': 'sc-cc-t', 
          'position': array([ 1.04211724,  3.82898322, 12.73732573]), 
          'normal': nan, 'indices': (1, 4), 'composition': None, 
          'subsurf_index': None, 'subsurf_element': None, 'label': '8OH', 
          'bonding_index': 40, 'bond_length': 0.66624136, 'adsorbate': 'CH3OH', 
          'fragment': 'OH', 'adsorbate_indices': (36, 37, 38, 39, 40, 41), 
          'occupied': 1, 'dentate': 2, 'fragment_indices': (40, 41)}, 
         {'site': 'bridge', 'surface': 'fcc211', 'morphology': 'sc-tc-h', 
          'position': array([ 1.04211724,  6.3816387 , 12.73732573]), 
          'normal': array([-0.33333333,  0.        ,  0.94280904]), 
          'indices': (2, 5), 'composition': None, 'subsurf_index': None, 
          'subsurf_element': None, 'label': '6CH3', 'bonding_index': 36, 
          'bond_length': 0.78070973, 'adsorbate': 'CH3OH', 'fragment': 'CH3', 
          'adsorbate_indices': (36, 37, 38, 39, 40, 41), 'occupied': 1, 
          'dentate': 2, 'fragment_indices': (36, 37, 38, 39)}]

    """

    def __init__(self, atoms, 
                 adsorption_sites=None, 
                 subtract_heights=None,
                 label_occupied_sites=False,
                 dmax=2.5, **kwargs):

        atoms = atoms.copy()
        ptp = np.ptp(atoms.positions[:, 2]) 
        if np.linalg.norm(atoms.cell[2]) - ptp < 10.:
            atoms.cell[2][2] = ptp + 10.

        self.atoms = atoms 
        self.positions = atoms.positions
        self.symbols = atoms.symbols
        self.ads_ids = [a.index for a in atoms if 
                        a.symbol in adsorbate_elements]
        if self.ads_ids:
            self.ads_atoms = atoms[self.ads_ids]
        else:
            self.ads_atoms = None
        self.cell = atoms.cell
        self.pbc = atoms.pbc

        self.subtract_heights = subtract_heights
        if subtract_heights is not None:
            self.subtract_heights = site_heights
            for k, v in subtract_heights.items():
                self.subtract_heights[k] = v      
        self.label_occupied_sites = label_occupied_sites
        self.dmax = dmax
        self.kwargs = {'allow_6fold': False, 'composition_effect': False,
                       'ignore_bridge_sites': False, 'label_sites': False} 
        self.kwargs.update(kwargs)
        if self.ads_ids:
            self.make_ads_neighbor_list()
            self.ads_adj_matrix = self.get_ads_connectivity() 
            self.identify_adsorbates()
        else:
            self.ads_list = []

        if adsorption_sites:
            sas = adsorption_sites
            for k in self.kwargs.keys():
                self.kwargs[k] = attrgetter(k)(sas)
        else:
            sas = SlabAdsorptionSites(atoms, **self.kwargs)    
        self.__dict__.update(self.kwargs)
        self.sas = sas
        self.slab_ids = sas.indices
        self.metals = sas.metals
        self.surf_ids = sas.surf_ids
        self.subsurf_ids = sas.subsurf_ids
        self.adj_matrix = sas.adj_matrix
        self.label_dict = sas.label_dict 
        self.hetero_site_list = deepcopy(sas.site_list)

        self.label_list = ['0'] * len(self.hetero_site_list)
        self.populate_hetero_site_list()

        if self.ads_ids:
            self.labels = self.get_occupied_labels() 
        else:
            self.labels = []

    def identify_adsorbates(self):

        G = nx.Graph()
        adscm = self.ads_adj_matrix

        # Cut all intermolecular H-H bonds except intramolecular        
        # H-H bonds in e.g. H2
        hids = [a.index for a in self.ads_atoms if a.symbol == 'H']
        for hi in hids:
            conns = np.where(adscm[hi] == 1)[0]
            hconns = [i for i in conns if self.ads_atoms.symbols[i] == 'H']
            if hconns and len(conns) > 1:
                adscm[hi,hconns] = 0

        if adscm.size != 0:
            np.fill_diagonal(adscm, 1)
            rows, cols = np.where(adscm == 1)

            edges = zip([self.ads_ids[row] for row in rows.tolist()], 
                        [self.ads_ids[col] for col in cols.tolist()])
            G.add_edges_from(edges)                        
            SG = (G.subgraph(c) for c in nx.connected_components(G))
 
            adsorbates = []
            for sg in SG:
                nodes = sorted(list(sg.nodes))
                adsorbates += [nodes]
        else:
            adsorbates = [self.ads_ids]

        self.ads_list = adsorbates

    def get_hetero_connectivity(self):
        """Get the adjacency matrix of slab + adsorbates."""

        nbslist = neighbor_shell_list(self.atoms, 0.3, neighbor_number=1)
        return get_adj_matrix(nbslist)                           

    def get_ads_connectivity(self):
        """Get the adjacency matrix for adsorbate atoms."""

        return get_adj_matrix(self.ads_nblist) 

    def get_site_connectivity(self):
        """Get the adjacency matrix for adsorption sites."""

        sl = self.hetero_site_list
        conn_mat = []
        for i, sti in enumerate(sl):
            conn_x = []
            for j, stj in enumerate(sl):
                overlap = len(set(sti['indices']).intersection(stj['indices']))
                if i == j:
                    conn_x.append(0.)
                elif overlap > 0:
                    if self.allow_6fold:         
                        if 'subsurf' in [sti['morphology'], stj['morphology']]: 
                            if overlap == 3:
                                conn_x.append(1.)
                            else:
                                conn_x.append(0.)
                        else:
                            conn_x.append(1.)
                    else:
                        conn_x.append(1.) 
                else:
                    conn_x.append(0.)
            conn_mat.append(conn_x)   

        return np.asarray(conn_mat) 

    def populate_hetero_site_list(self):
        """Find all the occupied sites, identify the adsorbate coverage
        of those sites and collect in a heterogeneous site list."""

        hsl = self.hetero_site_list
        ll = self.label_list
        ads_list = self.ads_list
        ndentate_dict = {} 

        for adsid in self.ads_ids:
            if self.symbols[adsid] == 'H':
                if [adsid] not in ads_list:
                    rest = [s for x in ads_list for s in x 
                            if (adsid in x and s != adsid)]
                    if not (self.symbols[rest[0]] == 'H' and len(rest) == 1):
                        continue

            adspos = self.positions[adsid]
            if self.subtract_heights is not None:
                _, dls = find_mic(np.asarray([s['position'] + s['normal'] * 
                                  self.subtract_heights[s['site']] - adspos 
                                  for s in hsl]), cell=self.cell, pbc=True)
                stid = np.argmin(dls) 
                st = hsl[stid]
                bl = get_mic(st['position'], adspos, self.cell,
                             return_squared_distance=True)**0.5
            else:
                _, bls = find_mic(np.asarray([s['position'] - adspos for s in hsl]),
                                  cell=self.cell, pbc=True)                                         
                stid = np.argmin(bls) 
                st, bl = hsl[stid], bls[stid]
            bl = round(bl, 8)
            if bl > self.dmax:
                continue

            adsids = next((l for l in ads_list if adsid in l), None)
            adsi = tuple(sorted(adsids))
            if 'occupied' in st:
                if bl >= st['bond_length']:
                    continue
                elif self.symbols[adsid] != 'H':
                    if adsi in ndentate_dict:
                        ndentate_dict[adsi] -= 1 
            st['bonding_index'] = adsid
            st['bond_length'] = bl

            symbols = ''.join(list(Formula(str(self.symbols[adsids]))))
            adssym = next((k for k, v in adsorbate_formulas.items() 
                           if v == symbols), None)
            if adssym is None:
                adssym = next((k for k, v in adsorbate_formulas.items()
                               if sorted(v) == sorted(symbols)), symbols)

            st['adsorbate'] = adssym
            st['fragment'] = adssym
            st['adsorbate_indices'] = adsi 
            if adsi in ndentate_dict:
                ndentate_dict[adsi] += 1
            else:
                ndentate_dict[adsi] = 1
            st['occupied'] = 1        

        # Get dentate numbers and coverage  
        self.n_occupied, n_surf_occupied, n_subsurf_occupied = 0, 0, 0
        for st in hsl:
            if 'occupied' not in st:
                st['bonding_index'] = st['bond_length'] = None
                st['adsorbate'] = st['fragment'] = None
                st['adsorbate_indices'] = None
                st['occupied'] = st['dentate'] = 0
                st['fragment_indices'] = None
                st['label'] = 0
                continue
            self.n_occupied += 1
            if st['morphology'] == 'subsurf':
                n_subsurf_occupied += 1
            else:
                n_surf_occupied += 1
            adsi = st['adsorbate_indices']
            if adsi in ndentate_dict:              
                st['dentate'] = ndentate_dict[adsi]
            else:
                st['dentate'] = 0
        self.coverage = n_surf_occupied / len(self.surf_ids)
        self.subsurf_coverage = n_subsurf_occupied / len(self.subsurf_ids)

        # Identify bidentate fragments and assign labels 
        self.multidentate_fragments = []
        self.monodentate_adsorbate_list = []
        self.multidentate_labels = []
        multidentate_adsorbate_dict = {}
        for j, st in enumerate(hsl):
            if st['occupied']:
                adssym = st['adsorbate']
                if st['dentate'] > 1:
                    bondid = st['bonding_index']
                    bondsym = self.symbols[bondid] 
                    conns = [self.ads_ids[k] for k in np.where(self.ads_adj_matrix[
                             self.ads_ids.index(bondid)] == 1)[0].tolist()]
                    hnnlen = len([i for i in conns if self.atoms[i].symbol == 'H'])
                    fsym = self.atoms[bondid].symbol
                    if hnnlen == 1:
                        fsym += 'H'
                    elif hnnlen > 1:
                        fsym += 'H{}'.format(hnnlen)
                    st['fragment'] = fsym 
                    flen = hnnlen + 1
                    adsids = st['adsorbate_indices']
                    ibond = adsids.index(bondid)
                    fsi = adsids[ibond:ibond+flen]
                    st['fragment_indices'] = fsi
                    if adsids not in multidentate_adsorbate_dict:
                        multidentate_adsorbate_dict[adsids] = adssym
                else:
                    st['fragment_indices'] = st['adsorbate_indices'] 
                    self.monodentate_adsorbate_list.append(adssym)

                if self.label_occupied_sites:
                    if st['label'] is None:
                        signature = [st['site'], st['morphology']]   
                        if self.composition_effect:
                            signature.append(st['composition'])
                        stlab = self.label_dict['|'.join(signature)]
                    else:
                        stlab = st['label']
                    label = str(stlab) + st['fragment']
                    st['label'] = label
                    ll[j] = label
                    if st['dentate'] > 1:                    
                        self.multidentate_fragments.append(label)
                        if bondid == adsids[0]:
                            mdlabel = str(stlab) + adssym
                            self.multidentate_labels.append(mdlabel)

        self.multidentate_adsorbate_list = list(multidentate_adsorbate_dict.values())
        self.adsorbate_list = self.monodentate_adsorbate_list + \
                              self.multidentate_adsorbate_list 

    def get_site(self, indices=None):
        """Get information of an adsorption site.
        
        Parameters
        ----------
        indices : list or tuple, default None
            The indices of the atoms that contribute to the site.
            Return a random site if the indices are not provided.   

        """

        if indices is None:
            st = random.choice(self.hetero_site_list)
        else:
            indices = indices if is_list_or_tuple(indices) else [indices]
            indices = tuple(sorted(indices))
            st = next((s for s in self.hetero_site_list if 
                       s['indices'] == indices), None)
        return st

    def get_sites(self, occupied_only=False):
        """Get information of all sites.
                                                                     
        Parameters                                                   
        ----------                                                   
        occupied_only : bool, default False
            Whether to only return occupied sites.

        """                                                          

        all_sites = self.hetero_site_list
        if occupied_only:
            all_sites = [s for s in all_sites if s['occupied']]
        return all_sites

    def get_adsorbates(self, adsorbate_species=None):
        """Get a list of tuples that contains each adsorbate (string)
        and the corresponding adsorbate indices (tuple) in ascending
        order.

        Parameters
        ----------
        adsorbate_species : list of strs, default None
            The available adsorbate species. If the adsorbate is not
            one of the available species, return the fragment instead.

        """

        adsorbates = []
        adsorbate_ids = set()
        for st in self.hetero_site_list:
            if st['occupied']:
                if adsorbate_species is not None:
                    if st['adsorbate'] not in adsorbate_species:
                        fragi = st['fragment_indices']
                        adsorbates.append((st['fragment'], fragi))
                        adsorbate_ids.update(fragi)
                        continue
                adsi = st['adsorbate_indices']
                if not set(adsi).issubset(adsorbate_ids):
                    adsorbates.append((st['adsorbate'], adsi))
                    adsorbate_ids.update(adsi)

        return sorted(adsorbates, key=lambda x: x[1])

    def get_fragments(self):
        """Get a list of tuples that contains each fragment (string)
        and the corresponding fragment indices (tuple) in ascending
        order."""

        fragments = []
        for st in self.hetero_site_list:
            if st['occupied']:
                fragi = st['fragment_indices']
                fragments.append((st['fragment'], fragi))

        return sorted(fragments, key=lambda x: x[1])        

    def make_ads_neighbor_list(self, dx=.2, neighbor_number=1):
        self.ads_nblist = neighbor_shell_list(self.ads_atoms, dx, 
                                              neighbor_number, mic=True)

    def get_occupied_labels(self, fragmentation=True):
        """Get a list of labels of all occupied sites. The label consists
        of a numerical part that represents site, and a character part
        that represents the occupying adsorbate.

        Parameters
        ----------
        fragmentation : bool, default True
            Whether to cut multidentate species into fragments. This ensures 
            that multidentate species with different orientations have 
            different labels.

        """

        if not self.label_occupied_sites:
            return self.atoms[self.ads_ids].get_chemical_formula(mode='hill')

        ll = self.label_list
        labs = [lab for lab in ll if lab != '0']
        if not fragmentation:
            mf = self.multidentate_fragments
            mdlabs = self.multidentate_labels
            c1, c2 = Counter(labs), Counter(mf)
            diff = list((c1 - c2).elements())
            labs = diff + mdlabs                   

        return sorted(labs)

    def get_graph(self, atom_wise=False,
                  fragmentation=True, 
                  subsurf_effect=False,
                  full_effect=False,
                  return_adj_matrix=False,
                  connect_dentates=True,
                  dx=0.5):                                         
        """Get the graph representation of the nanoparticle with adsorbates.

        Parameters
        ----------
        atom_wise : bool, default False
            Whether to treat each adsorbate as an atom-wise subgraph. If not, 
            treat each adsorbate fragment as a unity (molecule-wise).

        fragmentation : bool, default True
            Whether to cut multidentate species into fragments. This ensures 
            that multidentate species with different orientations have 
            different graphs.

        subsurf_effect : bool, default False
            Take subsurface atoms into consideration when genearting graph. 

        full_effect : bool, default False
            Take the whole catalyst into consideration when generating graph.

        return_adj_matrix : bool, default False
            Whether to return adjacency matrix instead of the networkx.Graph 
            object.

        connect_dentates : bool, default True
            Whether to add edges between the fragments of multidentate species.

        dx : float, default 0.5
            Buffer to calculate nearest neighbor pairs. Only relevent when
            atom_wise=True.

        """

        # Molecule-wise
        if not atom_wise:
            hsl = self.hetero_site_list
            hcm = self.adj_matrix.copy()
            if full_effect:
                surf_ids = self.slab_ids
            else:
                if subsurf_effect:
                    surf_ids = self.surf_ids + self.subsurf_ids
                else:
                    surf_ids = self.surf_ids
            surfhcm = hcm[surf_ids]
            symbols = self.symbols[surf_ids]
            nrows, ncols = surfhcm.shape       
 
            newrows, frag_list, adsi_list = [], [], []
            multi_conns = defaultdict(list)
            ohsl = [st for st in hsl if st['occupied']]
            for st in sorted(ohsl, key=lambda x: x['fragment_indices']):
                adsi = st['adsorbate_indices']
                if not fragmentation and st['dentate'] > 1: 
                    if st['bonding_index'] != adsi[0]:
                        continue
                si = st['indices']
                newrow = np.zeros(ncols)
                newrow[list(si)] = 1
                newrows.append(newrow)
 
                if fragmentation:                
                    frag_list.append(st['fragment'])
                    if st['dentate'] > 1 and connect_dentates:
                        bondid = st['bonding_index']
                        all_conns = [self.ads_ids[j] for j in np.where(self.ads_adj_matrix[
                                     self.ads_ids.index(bondid)] == 1)[0].tolist()]
                        conns = [c for c in all_conns if c not in st['fragment_indices']]
                        i = len(newrows) - 1
                        if i not in multi_conns:
                            multi_conns[bondid].append([i])
                        for k, v in multi_conns.items():
                            if k in conns:
                                multi_conns[k].append([i, v[0][0]])
                else:
                    frag_list.append(st['adsorbate'])
 
            links = []
            if newrows:
                surfhcm = np.vstack((surfhcm, np.asarray(newrows)))
                if multi_conns:
                    links = [sorted(c) for cs in multi_conns.values() for c in cs if len(c) > 1]
 
            shcm = surfhcm[:,surf_ids]
            if return_adj_matrix:
                if newrows:
                    dd = len(newrows)
                    small_mat = np.zeros((dd, dd))
                    if links:
                        for (i, j) in links:
                            small_mat[i,j] = 1
                            small_mat[j,i] = 1
                    return np.hstack((shcm, np.vstack((shcm[-dd:].T, small_mat))))
                else:
                    return shcm
 
            G = nx.Graph()               
            # Add nodes from fragment list
            G.add_nodes_from([(i, {'symbol': symbols[i]}) 
                               for i in range(nrows)] + 
                             [(j + nrows, {'symbol': frag_list[j]})
                               for j in range(len(frag_list))])
 
            # Add edges from surface adjacency matrix
            shcm = shcm * np.tri(*shcm.shape, k=-1)
            rows, cols = np.where(shcm == 1)
            edges = zip(rows.tolist(), cols.tolist())
            G.add_edges_from(edges)
            if links:
                for (i, j) in links:
                    G.add_edge(i + len(surf_ids), j + len(surf_ids))
        # Atom-wise
        else:
            nblist = neighbor_shell_list(self.atoms, dx=dx, 
                                         neighbor_number=1, mic=True)
            cm = get_adj_matrix(nblist)
            if full_effect:
                surf_ids = self.slab_ids
            else:
                if subsurf_effect:
                    surf_ids = self.surf_ids + self.subsurf_ids
                else:
                    surf_ids = self.surf_ids
            surf_ads_ids = sorted(surf_ids + self.ads_ids)
            shcm = cm[surf_ads_ids]
            symbols = self.symbols[surf_ads_ids]
            if return_adj_matrix:
                return shcm
 
            G = nx.Graph()                                                  
            G.add_nodes_from([(i, {'symbol': symbols[i]}) 
                              for i in range(len(symbols))])
            rows, cols = np.where(shcm == 1)
            edges = zip(rows.tolist(), cols.tolist())
            G.add_edges_from(edges)

        return G

#    def get_surface_bond_count_matrix(self, species):
#        hsl = self.hetero_site_list
#        cm = self.adj_matrix
#        atoms = self.atoms
#        numbers = atoms.numbers
#        symbols = atoms.symbols
#        specs = species
#        specs.sort(key=lambda x: atomic_numbers[x])
#        ncols = len(specs) + 1
#        sbcm = np.zeros((len(atoms), ncols))
#        for st in hsl:
#            frags = list(Formula(st['fragment']))
#            counts = Counter(frags)
#            for i in st['indices']:
#                for j, spec in enumerate(specs):
#                    sbcm[i,j] += counts[spec]
#        top_ids = self.surf_ids + self.subsurf_ids if \
#                  self.allow_6fold else self.surf_ids
#        for si in top_ids:
#            nbids = np.where(cm[si]==1)[0]
#            nbs = [symbols[i] for i in nbids]
#            nbcounts = Counter(nbs)
#            for j, spec in enumerate(specs):
#                sbcm[si,j] += nbcounts[spec]
#            sbcm[si,ncols-1] = numbers[si] 
#
#        return sbcm[top_ids]

    def get_coverage(self):
        """Get the adsorbate coverage (ML) of the surface."""

        return self.coverage

    def get_subsurf_coverage(self):
        """Get the adsorbate coverage (ML) of the subsurface."""

        return self.subsurf_coverage


def enumerate_occupied_sites(atoms, adsorption_sites=None,
                             surface=None, 
                             subtract_heights=None,
                             label_occupied_sites=False,                             
                             dmax=2.5, **kwargs):
    """A function that enumerates all occupied adsorption sites of
    the input atoms object. The function is generalized for both 
    periodic and non-periodic systems (distinguished by atoms.pbc).

    Parameters
    ----------
    atoms : ase.Atoms object
        Accept any ase.Atoms object. No need to be built-in.

    adsorption_sites : acat.adsorption_sites.ClusterAdsorptionSites \
        object or acat.adsorption_sites.SlabAdsorptionSites object, \
        default None
        The built-in adsorption sites class. If this is not provided, 
        the arguments for identifying adsorption sites can still be 
        passed in by **kwargs.

    surface : str, default None
        The surface type (crystal structure + Miller indices)
        If the structure is a periodic surface slab, this is required.
        If the structure is a nanoparticle, the function enumerates
        only the sites on the specified surface.

    subtract_heights : dict, default None
        A dictionary that contains the height to be subtracted from the 
        bond length when allocating a type of site to an adsorbate. 
        Default is to allocate the site that is closest to the adsorbate's 
        binding atom without subtracting height. Useful for ensuring the 
        allocated site for each adsorbate is consistent with the site to 
        which the adsorbate was added.                                                           

    label_occupied_sites : bool, default False
        Whether to assign a label to the occupied each site. The string 
        of the occupying adsorbate is concatentated to the numerical 
        label that represents the occupied site.

    dmax : float, default 2.5
        The maximum bond length (in Angstrom) between an atom and its
        nearest site to be considered as the atom being bound to the site.

    Example
    -------
    This is an example of enumerating all occupied sites on a truncated 
    octahedral nanoparticle:

        >>> from acat.adsorption_sites import ClusterAdsorptionSites
        >>> from acat.adsorbate_coverage import enumerate_occupied_sites
        >>> from acat.build import add_adsorbate_to_site
        >>> from ase.cluster import Octahedron
        >>> atoms = Octahedron('Ni', length=7, cutoff=2)
        >>> for atom in atoms:
        ...     if atom.index % 2 == 0:
        ...         atom.symbol = 'Pt'
        >>> atoms.center(vacuum=5.)
        >>> cas = ClusterAdsorptionSites(atoms, composition_effect=True,
        ...                              surrogate_metal='Ni') 
        >>> sites = cas.get_sites()
        >>> for s in sites:
        ...     if s['site'] == 'ontop':
        ...         add_adsorbate_to_site(atoms, adsorbate='OH', site=s)
        >>> sites = enumerate_occupied_sites(atoms, adsorption_sites=cas) 
        >>> print(sites[0])

    Output:

    .. code-block:: python

        {'site': 'ontop', 'surface': 'edge', 
         'position': array([ 6.76,  6.76, 12.04]), 
         'normal': array([-0.70710678, -0.70710678, -0.        ]), 
         'indices': (0,), 'composition': 'Pt', 'subsurf_index': None, 
         'subsurf_element': None, 'label': None, 'bonding_index': 201, 
         'bond_length': 1.8, 'adsorbate': 'OH', 'fragment': 'OH', 
         'adsorbate_indices': (201, 202), 'occupied': 1, 'dentate': 1, 
         'fragment_indices': (201, 202)}

    """

    if True not in atoms.pbc:
        cac = ClusterAdsorbateCoverage(atoms, adsorption_sites,
                                       subtract_heights,
                                       label_occupied_sites, 
                                       dmax, **kwargs)
        all_sites = cac.hetero_site_list
        if surface is not None:
            occupied_sites = [s for s in all_sites if s['surface'] 
                              == surface and s['occupied']]
        else:
            occupied_sites = [s for s in all_sites if s['occupied']]

    else:
        sac = SlabAdsorbateCoverage(atoms, surface, 
                                    adsorption_sites,
                                    subtract_heights,
                                    label_occupied_sites, 
                                    dmax, **kwargs)
        all_sites = sac.hetero_site_list
        occupied_sites = [s for s in all_sites if s['occupied']]

    return occupied_sites



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Comparator objects based on graph theory."""
from ..adsorbate_coverage import ClusterAdsorbateCoverage
from ..adsorbate_coverage import SlabAdsorbateCoverage
from ..utilities import neighbor_shell_list, get_adj_matrix
from ase.atoms import Atoms
from copy import deepcopy
import networkx as nx
import numpy as np


class AdsorptionGraphComparator(object):
    """Compares the graph of adsorbate overlayer + surface atoms and 
    returns True if they are isomorphic with node matches. Before checking 
    graph isomorphism by color refinement, a cheap label match is used to 
    reject graphs that are impossible to be isomorphic.

    The graphs (nx.Graph objects) can be quite costly to obtain every time 
    a graph is required (and disk intensive if saved), thus it makes sense 
    to get the graph along with e.g. the potential energy and save it in 
    atoms.info['data']['graph'].

    Parameters
    ----------
    adsorption_sites : acat.adsorption_sites.ClusterAdsorptionSites object \
        or acat.adsorption_sites.SlabAdsorptionSites object
        Provide the acat built-in adsorption sites class to accelerate the 
        pattern generation. Make sure all the structures have the same 
        atom indexing. 

    composition_effect : bool, default True
        Whether to consider sites with different elemental compositions as 
        different sites. It is recommended to set composition_effet=False 
        for monometallics.

    subtract_height : bool, default False
        Whether to subtract the height from the bond length when allocating
        a site to an adsorbate. Default is to allocate the site that is
        closest to the adsorbate's binding atom without subtracting height.
        Useful for ensuring the allocated site for each adsorbate is
        consistent with the site to which the adsorbate was added. 

    hmax : int, default 2
        Maximum number of iterations for color refinement.

    dmax : float, default 2.5
        The maximum bond length (in Angstrom) between an atom and its
        nearest site to be considered as the atom being bound to the site.

    """

    def __init__(self, adsorption_sites,  
                 composition_effect=True,
                 subtract_height=False,
                 hmax=2, dmax=2.5, **kwargs):
        
        self.adsorption_sites = adsorption_sites
        self.composition_effect = composition_effect
        self.subtract_height = subtract_height
        self.hmax = hmax
        self.dmax = dmax
        self.kwargs = kwargs
        self.comp = WLGraphComparator(hmax=self.hmax)

    def looks_like(self, a1, a2):
        isocheck = False
        if ('data' in a1.info and 'graph' in a1.info['data']) and (
        'data' in a2.info and 'graph' in a2.info['data']):
            isocheck = True
            G1 = a1.info['data']['graph']
            G2 = a2.info['data']['graph']
        else:
            sas = deepcopy(self.adsorption_sites)        
 
            if hasattr(sas, 'surface'):
                sas.update(a1, update_composition=self.composition_effect)
                sac1 = SlabAdsorbateCoverage(a1, sas, subtract_height=
                                             self.subtract_height, 
                                             label_occupied_sites=True, 
                                             dmax=self.dmax)
                sas.update(a2, update_composition=self.composition_effect)
                sac2 = SlabAdsorbateCoverage(a2, sas, subtract_height=
                                             self.subtract_height, 
                                             label_occupied_sites=True, 
                                             dmax=self.dmax)
            else:
                sas.update(a1, update_composition=self.composition_effect)
                sac1 = ClusterAdsorbateCoverage(a1, sas, subtract_height=
                                                self.subtract_height, 
                                                label_occupied_sites=True,
                                                dmax=self.dmax)
                sas.update(a2, update_composition=self.composition_effect)
                sac2 = ClusterAdsorbateCoverage(a2, sas, subtract_height=
                                                self.subtract_height, 
                                                label_occupied_sites=True,
                                                dmax=self.dmax)
            labs1 = sac1.get_occupied_labels(fragmentation=self.kwargs.get(
                                             'fragmentation', True))
            labs2 = sac2.get_occupied_labels(fragmentation=self.kwargs.get(
                                             'fragmentation', True))       
 
            if labs1 == labs2:
                isocheck = True 
                G1 = sac1.get_graph(self.kwargs)
                G2 = sac2.get_graph(self.kwargs)
        if isocheck:
            if self.comp.looks_like(G1, G2):
                return True

        return False


class WLGraphComparator(object):
    """Compares two structures (or graphs) based on the Weisfeiler-Lehman 
    subtree kernel (color refinement), as described in N. Shervashidze et 
    al., Journal of Machine Learning Research 2011, 12, 2539–2561. This 
    serves as a scalable solver for checking graph isomorphism of two 
    structures. 

    The graphs (nx.Graph objects) can be quite costly to obtain every time 
    a graph is required (and disk intensive if saved), thus it makes sense 
    to get the graph along with e.g. the potential energy and save it in 
    atoms.info['data']['graph'].

    Parameters
    ----------
    hmax : int, default 2
        Maximum number of iterations for color refinement.

    dx : float, default 0.5
        Buffer to calculate nearest neighbor pairs.

    tol : float, default 1e-5
        Tolerance of the discrepancy between K12 and sqrt(K11*K22)

    """

    def __init__(self, hmax=2, dx=.5, tol=1e-5):
        self.hmax = hmax
        self.dx = dx
        self.tol = tol

    def looks_like(self, a1, a2):
        if isinstance(a1, Atoms) and isinstance(a2, Atoms):
            if ('data' in a1.info and 'graph' in a1.info['data']) and (
            'data' in a2.info and 'graph' in a2.info['data']):
                a1 = a1.info['data']['graph']
                a2 = a2.info['data']['graph']
        d1 = WLGraphComparator.get_label_dict(a1, self.hmax, self.dx)
        d2 = WLGraphComparator.get_label_dict(a2, self.hmax, self.dx)
        d12 = {k: [d[k] if k in d else 0 for d in (d1, d2)] 
               for k in set(d1.keys()) | set(d2.keys())}
        X12 = np.asarray(list(d12.values()))

        # Compute the inner products
        k11 = sum(v**2 for v in d1.values())
        k22 = sum(v**2 for v in d2.values())
        k12 = X12[:,0] @ X12[:,1].T

        return abs(k12 - (k11 * k22)**0.5) <= self.tol 

    @classmethod
    def get_label_dict(cls, G, hmax, dx):                                                        
        d = {}
        if isinstance(G, Atoms):
            atoms = G.copy()
            symbols = atoms.symbols
            nblist = neighbor_shell_list(atoms, dx=dx, neighbor_number=1, 
                                         mic=(True in atoms.pbc))                     
            A = get_adj_matrix(nblist)
            N = len(A)
            G = nx.from_numpy_matrix(A)
        else:
            N = G.number_of_nodes()
            symbols = np.asarray([G.nodes[i]['symbol'] for i in range(N)], 
                                 dtype=object)

        nnlabs, neighbors = {}, {}
        isolates = []
        for i in range(N):
            lab0 = str(symbols[i])
            if lab0 in d:
                d[lab0] += 1.
            else:
                d[lab0] = 1.
            if hmax > 0:
                nnd = nx.single_source_shortest_path_length(G, i, cutoff=1)
                nns = [j for j, v in nnd.items() if v == 1]
                neighbors[i] = nns
                if len(nns) == 0:
                    isolates.append(i)
                    continue
                sorted_symbols = sorted(symbols[nns])
                lab1 = lab0 + ':' + ','.join(sorted_symbols)
                nnlabs[i] = lab1
                if lab1 in d:
                    d[lab1] += 1 
                else:
                    d[lab1] = 1
        if hmax > 1: 
            for k in range(1, hmax):
                nnnlabs = {}
                for i in range(N):
                    if i in isolates:
                        continue
                    nnlab = nnlabs[i]
                    nnnlab = ','.join(sorted(nnlabs[nn] for nn in neighbors[i]))
                    lab = nnlab + ':' * (k + 1) + nnnlab
                    nnnlabs[i] = lab
                    if lab in d:
                        d[lab] += 1
                    else:
                        d[lab] = 1
                nnlabs = nnnlabs 

        return d



from acat.build.adlayer import OrderedPatternGenerator as OPG
from acat.adsorption_sites import SlabAdsorptionSites
from ase.io import read
from ase.build import fcc111
from ase.visualize import view

# Initialize a pure Ni(111) slab
atoms = fcc111('Ni', (4, 4, 4), vacuum=5.)
for atom in atoms:
    if atom.index % 2 == 0:
        atom.symbol = 'Cu'

# Identify the surface adsorption sites (ignoring bridge sites)
sas = SlabAdsorptionSites(atoms, surface='fcc111',
                          ignore_bridge_sites=True,
                          surrogate_metal='Ni')

# Generate 50 unique adsorbate-alloy configurations with ordered
# adlayers, each consisting of at most 2 of the C, N, O adsorbates,
# and save in a trajectory file 'patterns_1.traj'
opg = OPG(atoms, adsorbate_species=['C', 'N', 'O'], surface='fcc111',
          max_species=2, adsorption_sites=sas, trajectory='patterns_1.traj')
opg.run(max_gen=50, unique=True)

# Visualize the output structures
images1 = read('patterns_1.traj', index=':')
view(images1)

# If you think the patterns are not 'ordered' enough, try providing
# a specific repeating distance, for example 5.026 Angstrom
opg = OPG(atoms, adsorbate_species=['C', 'N', 'O'], surface='fcc111',
          repeating_distance=5.026, max_species=2, adsorption_sites=sas, 
          trajectory='patterns_2.traj')
opg.run(max_gen=50, unique=True)

# Visualize the output structures
images2 = read('patterns_2.traj', index=':')
view(images2)

# If you want to generate very ordered well-known adlayer patterns,
# try acat.build.adlyer.special_coverage_pattern
from acat.build.adlayer import special_coverage_pattern as scp

# Generate the ordered adlayer pattern at 0.25 ML CO coverage
pattern = scp(atoms, adsorbate_species='CO', coverage=0.25, surface='fcc111')
view(pattern)

# Generate the ordered adlayer pattern at 0.5 ML CO coverage
pattern = scp(atoms, adsorbate_species='CO', coverage=0.5, surface='fcc111')
view(pattern)

# Generate the ordered adlayer pattern at 0.75 ML CO coverage
pattern = scp(atoms, adsorbate_species='CO', coverage=0.75, surface='fcc111')
view(pattern)

# Generate the ordered adlayer pattern at 1 ML CO coverage
pattern = scp(atoms, adsorbate_species='CO', coverage=1., surface='fcc111')
view(pattern)




from acat.adsorption_sites import ClusterAdsorptionSites
from acat.adsorbate_coverage import ClusterAdsorbateCoverage
from acat.build import add_adsorbate_to_site
from ase.cluster import Octahedron
from ase.visualize import view

# Initialize a pure Ni truncated octahedral nanoparticle
atoms = Octahedron('Ni', length=7, cutoff=2)
atoms.center(vacuum=5.)

# Generate a NiPt nanoalloy by changing some Ni to Pt
for atom in atoms:
    if atom.index % 2 == 0:
        atom.symbol = 'Pt'

# Identify all adsorption sites on the nanoalloy
cas = ClusterAdsorptionSites(atoms, composition_effect=True,
                             surrogate_metal='Ni') 

# Generate a Ni-Pt nanoalloy covered by CO at all fcc and 4fold sites
# (as an example of the input adsorbate-alloy structure one would provide)
sites = cas.get_sites()
for site in sites:
    if site['site'] in ['fcc', '4fold']:
        add_adsorbate_to_site(atoms, adsorbate='CO', site=site)

# Visualize the structure
view(atoms)

# Get information about the surface adsorbate coverage
cac = ClusterAdsorbateCoverage(atoms, adsorption_sites=cas,
                               label_occupied_sites=True)

# Print all occupied sites
occupied_sites = cac.get_sites(occupied_only=True)
for i, site in enumerate(occupied_sites):
    print('Occupied site {0}: {1}'.format(i, site))

# Calculate the surface coverage (ML)
print('Surface coverage: {0}'.format(cac.get_coverage()))

# Get information of all adsorbates
adsorbates_info = cac.get_adsorbates()
for i, ads_info in enumerate(adsorbates_info):
    ads_species, ads_indices = ads_info
    print('Adsorbate {0}: {1} with atomic indices of {2}'.format(i, ads_species, ads_indices))




from acat.build.ordering import RandomOrderingGenerator as ROG
from ase.build import bulk
from ase.io import read
from ase.visualize import view

# Initialize a pure Ni bulk structure
atoms = bulk('Ni', 'fcc')
atoms *= (6, 6, 6)

# Generate 50 NiPtAu bulk alloys with random chemical 
# orderings and save in a trajectory file 'orderings_1.traj'
rog = ROG(atoms, elements=['Ni','Pt','Au'], 
          trajectory='orderings_1.traj')
rog.run(num_gen=50)

# Visualize the output structures 
images1 = read('orderings_1.traj', index=':')
view(images1)

# Now generate 50 unique NiPtAu bulk alloys with the composition 
# of 1:1:1 and save in a trajectory file 'orderings_2.traj'
rog = ROG(atoms, elements=['Ni','Pt','Au'], 
          composition={'Ni': 1,'Pt': 1,'Au': 1},          
          trajectory='orderings_2.traj')
rog.run(num_gen=50)

# Visualize the output structures
images2 = read('orderings_2.traj', index=':')
view(images2)




import setuptools
import os

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

thelibFolder = os.path.dirname(os.path.realpath(__file__))
requirementPath = thelibFolder + '/requirements.txt'
install_requires = [] 
if os.path.isfile(requirementPath):
    with open(requirementPath) as f:
        install_requires = f.read().splitlines()

exec(open('acat/version.py').read())
setuptools.setup(
    name='acat', 
    version=__version__,
    author='Shuang Han',
    author_email='hanshuangshiren@gmail.com',
    description='Alloy Catalysis Automated Toolkit',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.com/asm-dtu/acat',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Operating System :: POSIX',
    ],
    install_requires=install_requires,
    python_requires='>=3.6',
)
